﻿--[[	*** DataStore_Talents ***
Written by : Thaoky, EU-Marécages de Zangar
June 23rd, 2009
--]]
DataStore_Talents = LibStub("AceAddon-3.0"):NewAddon("DataStore_Talents", "AceConsole-3.0", "AceEvent-3.0")

local addon = DataStore_Talents

-- TODO: 
	-- add glyphs
	-- see how to work around the tab order issue
	-- add support for hunter pets' talent trees

local THIS_ACCOUNT = "Default"
local NUM_GLYPH_SLOTS = 6

local AddonDB_Defaults = {
	global = {
		['*'] = {							-- Account: "Default" or user defined for other accounts
			['*'] = {						-- Realm
				Characters = {
					['*'] = {				-- Character Name
						ActiveTalents = nil,		-- 1 for primary, 2 for secondary
						Class = nil,				-- englishClass
						PointsSpent = nil,		-- "51,5,15 ...	" 	3 numbers for primary spec, 3 for secondary, comma separated
						TalentTrees = {
							['*'] = {		-- "Fire|2"	= Mage Fire tree, secondary
								['*'] = 0
							}
						},
						Glyphs = {},
					}
				}
			}
		}
	}
}

-- This table saved reference data required to rebuild a talent tree for a class when logged in under another class.
-- The API does not provide that ability, but saving and reusing is fine
local ReferenceDB_Defaults = {
	global = {
		['*'] = {							-- "englishClass" like "MAGE", "DRUID" etc..
			Order = nil,
			Trees = {
				['*'] = {						-- tree name
					icon = nil,
					background = nil,
					list = { ['*'] = nil },		-- name, icon, max rank etc..for talent x in this tree
					prereqs = { ['*'] = nil }	-- prerequisites
				},
			}
		},
	}
}


function addon:OnInitialize()
	addon.db = LibStub("AceDB-3.0"):New("DataStore_TalentsDB", AddonDB_Defaults)
	addon.ref = LibStub("AceDB-3.0"):New("DataStore_TalentRefDB", ReferenceDB_Defaults)
end

function addon:OnEnable()
	local realm = GetRealmName()
	local player = UnitName("player")
	
	self.ThisAccount = addon.db.global[THIS_ACCOUNT]
	self.ThisRealm = self.ThisAccount[realm]
	self.ThisCharacter = self.ThisRealm.Characters[player]

	self:RegisterEvent("PLAYER_ALIVE")
	self:RegisterEvent("PLAYER_TALENT_UPDATE")
	self:RegisterEvent("GLYPH_ADDED", self.ScanGlyphs)
	self:RegisterEvent("GLYPH_REMOVED", self.ScanGlyphs)
	self:RegisterEvent("GLYPH_UPDATED", self.ScanGlyphs)
end

function addon:OnDisable()
	self:UnregisterEvent("PLAYER_ALIVE")
	self:UnregisterEvent("PLAYER_TALENT_UPDATE")
	self:UnregisterEvent("GLYPH_ADDED")
	self:UnregisterEvent("GLYPH_REMOVED")
	self:UnregisterEvent("GLYPH_UPDATED")
end

-- *** Utility functions ***

-- *** Scanning functions ***
local TALENT_ICON_PATH = "Interface\\Icons\\"
local BACKGROUND_PATH = "Interface\\TalentFrame\\"

--- see ScanTalentReference() for an explanation on the purpose of this table
local LocaleExceptions = {}
	
if GetLocale() == "enUS" then
	LocaleExceptions["Elemental Combat"] = "Elemental"
	LocaleExceptions["Shadow Magic"] = "Shadow"
end

local function ScanTalents()

	local level = UnitLevel("player")
	if not level or level < 10 then return end		-- don't scan anything for low level characters

	local char = addon.ThisCharacter
	local _, englishClass = UnitClass("player")
	
	char.ActiveTalents = GetActiveTalentGroup()		-- returns 1 or 2
	char.Class = englishClass
	
	wipe(char.TalentTrees)
	
	local points = {}
	
	for specNum = 1, 2 do												-- primary and secondary specs
		for tabNum = 1, GetNumTalentTabs() do						-- all tabs
			local name, _, pointsSpent = GetTalentTabInfo( tabNum, nil, nil, specNum );
			table.insert(points, pointsSpent)
			
			for talentNum = 1, GetNumTalents(tabNum) do			-- all talents
				local _, _, _, _, currentRank = GetTalentInfo( tabNum, talentNum, nil, nil, specNum )

				char.TalentTrees[name .."|" .. specNum][talentNum] = currentRank
			end
		end
	end
	
	char.PointsSpent = table.concat(points, ",")
end

local function ScanTalentReference()

	local level = UnitLevel("player")
	if not level or level < 10 then return end		-- don't scan anything for low level characters

	local ref = addon.ref.global
	local _, class = UnitClass("player")		-- we need the englishClass
	local order = {}									-- order of the talent tabs
	
	-- first talent tree, gather reference + user specific
	for tabNum = 1, GetNumTalentTabs() do
		local talentTabName, _, _, fileName = GetTalentTabInfo( tabNum, nil, nil, 1 );
		order[tabNum] = talentTabName
		
		local ti = ref[class].Trees[talentTabName]		-- ti for talent info

		ti.background = fileName
			
		for talentNum = 1, GetNumTalents(tabNum) do
			local nameTalent, iconPath, tier, column, _, maximumRank = GetTalentInfo(tabNum, talentNum, nil, nil, 1 )

			-- all paths start with this prefix, let's hope blue does not change this :)
			-- saves a lot of memory not to keep the full path for each talent (about 16k in total for all classes)
			iconPath = string.gsub(iconPath, TALENT_ICON_PATH, "")
			
			local link = GetTalentLink(tabNum, talentNum)
			local id = tonumber(link:match("talent:(%d+)"))
			
			ti.list[talentNum] = id .. "|" .. nameTalent .. "|" .. iconPath .. "|" .. tier .. "|" ..  column .. "|" .. maximumRank
			
			prereqTier, prereqColumn = GetTalentPrereqs(tabNum, talentNum)		-- talent prerequisites
			if prereqTier and prereqColumn then
				ti.prereqs[talentNum] = prereqTier .. "|" .. prereqColumn
			end
		end
	end
	
	-- save the order of talent tabs, this is necessary because the order of talent tabs is not the same as that of spell tabs in all languages/classes
	-- it is fine in enUS, but not in frFR (druid at least did not match)
	ref[class]["Order"] = table.concat(order, ",")
	
	for i = 2, 4 do
		local name, icon = GetSpellTabInfo(i)		-- skip spell tab 1, it's the general tab
		
		-- the icon may be nil on a low level char. 
		-- Example : rogue lv 2
			-- GetSpellTabInfo(1) returns the General tab
			-- GetSpellTabInfo(2) returns the Assassination tab
			-- GetSpellTabInfo(3) returns the Combat tab
			-- GetSpellTabInfo(4) returns nil, instead of Subtelty
		if name and icon then
			-- in addition to having different order between spell tabs & talent tabs, the names may not match. Most of the time they do, but in case they don't, use the excpetion table
			name = LocaleExceptions[name] or name
		
		
			local ti = ref[class].Trees[name]		-- ti for talent info
			ti.icon = string.gsub(icon, TALENT_ICON_PATH, "")
		end
	end
end

local function ScanGlyphs()
	-- GLYPHTYPE_MAJOR = 1;
	-- GLYPHTYPE_MINOR = 2;

	--		1
	--	3		5
	--	6		4
	--		2

	local glyphs = addon.ThisCharacter.Glyphs
	wipe(glyphs)
	
	local enabled, glyphType, spell, icon, glyphID
	local link, index
	
	for specNum = 1, 2 do
		for i = 1, NUM_GLYPH_SLOTS do
			index = ((specNum - 1) * NUM_GLYPH_SLOTS) + i
	      
		   enabled, glyphType, spell, icon = GetGlyphSocketInfo(i, specNum)
			link = GetGlyphLink(i, specNum)
			if link then
				_, glyphID = link:match("glyph:(%d+):(%d+)")
			end
			
			glyphID = glyphID or 0
			glyphType = glyphType or 0
			enabled = enabled or 0
			spell = spell or ""
			icon = icon or ""
			
			glyphs[index] = enabled .."|" .. glyphType .. "|" .. spell .. "|" .. icon .. "|" .. glyphID
		end
	end
end

-- *** DB functions / TALENTS ***
function addon:GetCharacter(name, realm, account)
	-- assert(type(name) == "string")
	-- assert(type(realm) == "string")
	-- assert(type(account) == "string")
	
	-- Usage: 
	-- 	local c = addon:GetCharacter(char, realm, account)
	--	all 3 parameters default to current player, realm or account
	name = name or UnitName("player")
	realm = realm or GetRealmName()
	account = account or THIS_ACCOUNT
	
	return addon.db.global[account][realm].Characters[name]
end

function addon:DeleteCharacter(name, realm, account)
	name = name or UnitName("player")
	realm = realm or GetRealmName()
	account = account or THIS_ACCOUNT

	addon.db.global[account][realm].Characters[name] = nil
end

function addon:GetReferenceTable()
	return addon.ref.global
end

function addon:GetClassReference(class)
	assert(type(class) == "string")
	return addon.ref.global[class]
end

function addon:GetTreeReference(class, tree)
	assert(type(class) == "string")
	assert(type(tree) == "string")
	
	return addon.ref.global[class].Trees[tree]
end

function addon:IsClassKnown(class)
	class = class or ""	-- if by any chance nil is passed, trap it to make sure the function does not fail, but returns nil anyway
	
	local ref = addon:GetClassReference(class)
	if ref.Order then		-- if the Order field is not nil, we have data for this class
		return true
	end
end

function addon:GetClassTrees(class)
	assert(type(class) == "string")
	
	local ref = addon:GetClassReference(class)
	local order = ref.Order
	if order then
		return order:gmatch("([^,]+)")
	end
	-- to do, add a return value that does not require validity testing by the caller
end

--[[	*** Sample code ***
	local DS = DataStore_Talents

	local character = DS:GetCharacter()
	local class = "MAGE"

	for v in DS:GetClassTrees(class) do
		print(v)
	end
--]]

function addon:GetTreeInfo(class, tree)
	local t = addon:GetTreeReference(class, tree)
	
	if t then
		return TALENT_ICON_PATH..t.icon, BACKGROUND_PATH .. t.background
	end	
end

function addon:GetTreeNameByID(class, id)
	-- returns the name of tree "id" for a given class
	assert(type(class) == "string")
	
	local index = 1
	for name in addon:GetClassTrees(class) do
		if index == id then
			return name
		end
		index = index + 1
	end
end

function addon:GetTalentLink(id, rank, name)
	return format("|cff4e96f7|Htalent:%s:%s|h[%s]|h|r", id, (rank-1), name)
end

function addon:GetNumTalents(class, tree)
	-- returns the number of talents in a given tree
	local t = addon:GetTreeReference(class, tree)

	if t then
		return #t.list
	end
end

function addon:GetTalentInfo(class, tree, index)
	local t = addon:GetTreeReference(class, tree)
	local talentInfo = t.list[index]
	
	if not talentInfo then return end
	
	local id, name, icon, tier, column, maximumRank	= strsplit("|", talentInfo)
	
	return tonumber(id), name, TALENT_ICON_PATH..icon, tonumber(tier), tonumber(column), tonumber(maximumRank)
end

--[[	*** Sample code ***
	local DS = DataStore_Talents

	local character = DS:GetCharacter()
	local class = "MAGE"
	local tree = "Fire"
	local spec = 1  -- primary spec

	for i = 1, DS:GetNumTalents(class, tree) do
		local id, name = DS:GetTalentInfo(class, tree, i)
		print(id .. ": " ..name)

		local rank = DS:GetTalentRank(DS:GetCharacter(), tree, spec, i)
		print(rank)

		local link = DS:GetTalentLink(id, rank, name)
		print(link)
	end
--]]

function addon:GetTalentRank(character, tree, specNum, index)
	assert(type(character) == "table")		-- this is the pointer to a character table, obtained through addon:GetCharacter()

	return character.TalentTrees[tree .. "|" .. specNum][index]
end

function addon:GetNumPointsSpent(character, tree, specNum)
	assert(type(character) == "table")		-- this is the pointer to a character table, obtained through addon:GetCharacter()
	
	local index = 1
	for treeName in addon:GetClassTrees(character.Class) do
		if treeName == tree then
			break
		end
		index = index + 1
	end
	
	if index == 4 then return end				-- = 4 means tree was not found
	
	index = index + ((specNum-1) * 3)
	
	return select(index, strsplit(",", character.PointsSpent)) or 0
end

function addon:GetTalentPrereqs(class, tree, index)
	local t = addon:GetTreeReference(class, tree)
	local prereq = t.prereqs[index]
	
	if prereq then
		local prereqTier, prereqColumn = strsplit("|", prereq)
		return tonumber(prereqTier), tonumber(prereqColumn)
	end
end

-- *** DB functions / GLYPHS ***

function addon:GetGlyphInfo(character, specNum, index)
	assert(type(character) == "table")		-- this is the pointer to a character table, obtained through addon:GetCharacter()
	
	index = ((specNum - 1) * NUM_GLYPH_SLOTS) + index		-- 4th glyph = 4 for spec 1, 10 for spec 2
	local glyph = character.Glyphs[index]
	local enabled, glyphType, spell, icon, glyphID = strsplit("|", glyph)

	return tonumber(enabled), tonumber(glyphType), tonumber(spell), icon, tonumber(glyphID)	
end

function addon:GetGlyphLink(id, spell, glyphID)
	local name = GetSpellInfo(spell)
	return format("|cff66bbff|Hglyph:2%s:%s|h[%s]|h|r", id, glyphID, name)
end


-- *** EVENT HANDLERS ***
function addon:PLAYER_ALIVE()
	ScanTalents()
	ScanTalentReference()
	ScanGlyphs()
end

function addon:PLAYER_TALENT_UPDATE()
	ScanTalents()
end

function addon:ScanGlyphs()
	ScanGlyphs()
end
